/*-
 * Copyright (c) 2011 Marius Strobl <marius@smtpd.org>
 * All rights reserved.  The same license as below applies.
 *
 * Basic device handling based on the UART driver tester available from:
 * http://p4db.freebsd.org/fileLogView.cgi?FSPC=//depot/projects/uart/dev/uart/uarttest.c
 * and marked:
 *
 * Copyright (c) 2003 Marcel Moolenaar
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: bert.c,v 1.10 2012/03/14 16:54:16 marius Exp $
 */

/*
 * Bit error rate tester (BERT) for serial communication via UARTs and
 * optionally Yamar SIG60
 */

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>

#include "bert_endian.h"
#include "sig60.h"

#define	AS_DCE		1
#define	AS_DTE		2

#define	SIG60_VAL_CR0_DEFAULT						\
	(SIG60_CR0_LWUM_EN | SIG60_CR0_ASLEEP_DIS |			\
	SIG60_CR0_AWUM_DIS | SIG60_CR0_LOOP_DIS)

#define	SIG60_VAL_CR1_DEFAULT	SIG60_CR1_F_10500_13000

#define	SIG60_HDC	TIOCM_DTR

#define	SIG60_SPEED	19200

#define	PRBS_IV		0x1

static void bail(int status);
static u_int bitcount(uint32_t val);
static void onsignal(int sig);
static uint32_t prbs(void);
static void sethow(int h);
static int sig60_read(uint8_t reg, uint8_t *val);
static int sig60_write(uint8_t reg, uint8_t val);
static void usage(void);

static uint32_t lfsr;

static uint32_t octets_recv;		/* octets received total */
static uint32_t octets_recv_berr;	/* octets received bit errors */

static struct termios tfd, tstdin;
static int fd;
static int how;
static const char* progname;

static void
bail(int status)
{

	tcsetattr(STDIN_FILENO, TCSANOW, &tstdin);
	tcsetattr(fd, TCSADRAIN, &tfd);
	close(fd);
	exit(status);
}

static void
onsignal(int sig)
{

	/* Von Osten kommt's geflogen! */
	switch (sig) {
	case SIGINT:
		bail(0);
		break;
	}
}

/*
 * Obtained from:
 * http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
 */
static u_int
bitcount(uint32_t val)
{

	val = val - ((val >> 1) & 0x55555555);
	val = (val & 0x33333333) + ((val >> 2) & 0x33333333);
	return ((((val + (val >> 4)) & 0xf0f0f0f) * 0x1010101) >> 24);
}

/*
 * PRBS31 generator according to IEEE 802.3-2008 49.2.8
 * polynomial: G(x) = 1 + x^28 + x^31, output inverted
 */
static uint32_t
prbs(void)
{

	lfsr = (lfsr >> 1) ^ (-(lfsr & (uint32_t)1) &
	    (((uint32_t)1 << 30) | ((uint32_t)1 << 27)));
	return (~lfsr);
}

static void
dte(void)
{
	char buf[sizeof(uint32_t)];
	size_t count;
	ssize_t wc;

	lfsr = PRBS_IV;

	for (;;) {
		bert_le32enc(buf, prbs());
		count = 0;
		while (count < sizeof(buf)) {
			wc = write(fd, buf + count, sizeof(buf) - count);
			if (wc > 0)
				count += wc;
		}
	}
}

static void
dce(void)
{
	char buf[sizeof(uint32_t)];
	struct timeval stv, tv;
	fd_set fdset;
	size_t count;
	ssize_t rc;
	double ber, stime;
	int locked;
	uint32_t recv, start;

	stime = 0;
	locked = 0;
	lfsr = PRBS_IV;
	start = prbs();
	stv.tv_sec = 0;
	stv.tv_usec = 0;

	for (;;) {
		count = 0;
		while (count < sizeof(buf)) {
			FD_ZERO(&fdset);
			FD_SET(STDIN_FILENO, &fdset);
			rc = select(STDIN_FILENO + 1, &fdset, NULL, NULL,
			    &stv);
			if (rc != 0) {
				(void)getc(stdin);
				if (locked == 0)
					printf("start not locked\n");
				else {
					(void)gettimeofday(&tv, NULL);
					ber = octets_recv_berr == 0 ? 0 :
					    (double)octets_recv_berr /
					    (octets_recv * 8);
					printf("running [s]: %.2f octets "
					    "total: %d bit errors: %d BER: "
					    "%s%.2f%s\n", tv.tv_sec +
					    tv.tv_usec * 1e-6 - stime,
					    octets_recv, octets_recv_berr,
					    ber == 0 ? "" : "10^(",
					    ber == 0 ? 0 : log10(ber),
					    ber == 0 ? "" : ")");
				}
			}
			rc = read(fd, buf + count, sizeof(buf) - count);
			if (rc > 0)
				count += rc;
		}
		recv = bert_le32dec(buf);
		if (recv == start) {
			lfsr = PRBS_IV;
			(void)prbs();
			(void)gettimeofday(&tv, NULL);
			stime = tv.tv_sec + tv.tv_usec * 1e-6;
			octets_recv = octets_recv_berr = 0;
			locked++;
			printf("start locked\n");
		} else {
			if (locked == 0)
				continue;
			octets_recv += count;
			octets_recv_berr += bitcount(recv & ~prbs());
		}
	}
}

static int
sig60_read(uint8_t reg, uint8_t *val)
{
	struct timespec ts;
	ssize_t count;
	int tries;

	reg = SIG60_READ(reg);
	count = write(fd, &reg, sizeof(reg));
	if (count <= 0) {
		perror("cannot write to SIG60");
		return (-1);
	}
	tcdrain(fd);
	ts.tv_sec = 0;
	ts.tv_nsec = 10 * 1000; /* 10 ms */
	for (tries = 10; tries > 0; tries--) {
		(void)nanosleep(&ts, NULL);
		count = read(fd, val, sizeof(*val));
		if (count > 0)
			return (0);
	}
	fprintf(stderr, "timeout reading from SIG60\n");
	return (-1);
}

static int
sig60_write(uint8_t reg, uint8_t val)
{
	size_t count;
	ssize_t wc;
	uint8_t data[2];

	data[0] = SIG60_WRITE(reg);
	data[1] = val;
	for (count = 0; count < sizeof(data);) {
		wc = write(fd, (caddr_t)&data + count, sizeof(data));
		if (wc <= 0) {
			perror("cannot write to SIG60");
			return (-1);
		}
		count += wc;
	}
	tcdrain(fd);
	return (0);
}

int
main(int argc, char *argv[])
{
	char buf[PATH_MAX];
	struct termios t;
	struct timespec ts;
	speed_t sp;
	int ch, error, f1, init, sig;
	uint8_t cfg, val;

	progname = argv[0];

	f1 = init = 0;
	sp = 9600;	/* Any speed that we expect to work for all UARTs */
	while ((ch = getopt(argc, argv, "cfis:t")) != -1) {
		switch (ch) {
		case 'c':	/* DCE mode */
			sethow(AS_DCE);
			break;
		case 'f':	/* set SIG60_CR0_F1NF0 */
			f1 = 1;
			break;
		case 'i':	/* init SIG60 */
			init = 1;
			break;
		case 's':	/* Line speed */
			sp = atoi(optarg);
			break;
		case 't':	/* DTE mode */
			sethow(AS_DTE);
			break;
		default:
			usage();
			/* NOTREACHED */
		}
	}

	if (how == 0) {
		usage();
		/* NOTREACHED */
	}

	argc -= optind;
	argv += optind;

	if (argc != 1) {
		usage();
		/* NOTREACHED */
	}

	tcgetattr(STDIN_FILENO, &tstdin);
	t = tstdin;
	/* Turn canonical mode and echo off. */
	t.c_lflag &= ~(ICANON | ECHO);
	tcsetattr(STDIN_FILENO, TCSANOW, &t);
	if (tcsetattr(STDIN_FILENO, TCSANOW, &t) == -1) {
		perror("cannot disable canonical mode and echo");
		exit(1);
	}

	fd = open(argv[0], O_RDWR | O_NONBLOCK);
	if (fd == -1) {
		if (errno == ENOENT) {
			snprintf(buf, PATH_MAX, "/dev/%s", argv[0]);
			fd = open(buf, O_RDWR | O_NONBLOCK);
			if (fd == -1)
				fprintf(stderr, "opening `%s' and `%s': %s\n",
				    argv[0], buf, strerror(errno));
		} else
			fprintf(stderr, "opening `%s': %s\n", argv[0],
			    strerror(errno));
	}
	if (fd == -1) {
		tcsetattr(STDIN_FILENO, TCSANOW, &tstdin);
		exit(1);
	}

	tcgetattr(fd, &tfd);
	t = tfd;
	cfmakeraw(&t);
	cfsetspeed(&t, init == 0 ? sp : SIG60_SPEED);
	/* Ignore modem status lines. */
	t.c_cflag |= CLOCAL;
	if (tcsetattr(fd, TCSAFLUSH | TCSANOW, &t) == -1) {
		perror("cannot set line attributes");
		tcsetattr(STDIN_FILENO, TCSANOW, &tstdin);
		close(fd);
		exit(1);
	}

	if (signal(SIGINT, onsignal) == SIG_ERR) {
		perror("cannot set SIGINT handler");
		bail(1);
		/* NOTREACHED */
	}

	if (init != 0) {
		/* Initialize a SIG60. */

		ts.tv_sec = 0;
		ts.tv_nsec = 10 * 1000; /* 10 ms */
		if (ioctl(fd, TIOCMGET, &sig) == -1) {
			perror("cannot get line state");
			bail(1);
			/* NOTREACHED */
		}
		/* set SIG60_HDC low */
		sig |= SIG60_HDC;
		if (ioctl(fd, TIOCMSET, &sig) == -1) {
			perror("cannot set line state");
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		/* set SIG60_CR0 */
		cfg = SIG60_VAL_CR0_DEFAULT;
		if (f1 != 0)
			cfg |= SIG60_CR0_F1NF0;
		error = sig60_write(SIG60_CR0, cfg);
		if (error < 0) {
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		error = sig60_read(SIG60_CR0, &val);
		if (error < 0) {
			bail(1);
			/* NOTREACHED */
		}
		if ((val & ~SIG60_CR0_IFER) != cfg) {
			fprintf(stderr,
			    "SIG60_CR0 mismatch (wrote 0x%x got 0x%x)\n", cfg,
			    val);
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		/* set SIG60_CR1 */
		cfg = SIG60_VAL_CR1_DEFAULT;
		switch(sp) {
		case 9600:
		case 115200:
			cfg |= SIG60_CR1_BR_9K6_115K2;
			break;
		case 19200:
			cfg |= SIG60_CR1_BR_19K2;
			break;
		case 38400:
			cfg |= SIG60_CR1_BR_38K4;
			break;
		case 57600:
			cfg |= SIG60_CR1_BR_57K6;
			break;
		default:
			fprintf(stderr, "not today\n");
			bail(1);
			/* NOTREACHED */
		}
		error = sig60_write(SIG60_CR1, cfg);
		if (error < 0) {
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		error = sig60_read(SIG60_CR1, &val);
		if (error < 0) {
			bail(1);
			/* NOTREACHED */
		}
		if (val != cfg) {
			fprintf(stderr,
			    "SIG60_CR1 mismatch (wrote 0x%x got 0x%x)\n", cfg,
			    val);
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		/* set SIG60_HDC high */
		sig &= ~SIG60_HDC;
		if (ioctl(fd, TIOCMSET, &sig) == -1) {
			perror("cannot set line state");
			bail(1);
			/* NOTREACHED */
		}
		(void)nanosleep(&ts, NULL);
		printf("SIG60 intialized\n");

		cfsetspeed(&t, sp);
		if (tcsetattr(fd, TCSAFLUSH | TCSANOW, &t) == -1) {
			perror("cannot set line attributes");
			bail(1);
			/* NOTREACHED */
		}
	}

	printf("Line speed is %u bps\n", sp);

	switch (how) {
	case AS_DCE:
		dce();
		break;
	case AS_DTE:
		dte();
		break;
	default:
		fprintf(stderr, "not today\n");
		bail(1);
		/* NOTREACHED */
	}

	return (0);
}

static void
sethow(int h)
{

	if (how != 0) {
		usage();
		/* NOTREACHED */
	}
	how = h;
}

static void
usage(void)
{

	fprintf(stderr, "usage: %s -[c|f|i|t] [-s speed] <device>\n",
	    progname);
	exit(1);
}
